"use strict";
this.haketilo_exports = this.haketilo_exports || {};
this.haketilo_exports["background/webrequest.js"] = {};

(function() {
var globalThis = this.haketilo_this
{
/**
 * This file is part of Haketilo.
 *
 * Function: Modify HTTP traffic usng webRequest API.
 *
 * Copyright (C) 2021, 2022 Wojtek Kosior <koszko@koszko.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * As additional permission under GNU GPL version 3 section 7, you
 * may distribute forms of that code without the copy of the GNU
 * GPL normally required by section 4, provided you include this
 * license notice and, in case of non-source distribution, a URL
 * through which recipients can access the Corresponding Source.
 * If you modify file(s) with this exception, you may extend this
 * exception to your version of the file(s), but you are not
 * obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * As a special exception to the GPL, any HTML file which merely
 * makes function calls to this code, and for that purpose
 * includes it by reference shall be deemed a separate work for
 * copyright law purposes. If you modify this code, you may extend
 * this exception to your version of the code, but you are not
 * obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * I, Wojtek Kosior, thereby promise not to sue for violation of this file's
 * license. Although I request that you do not make use of this code in a
 * proprietary program, I am not going to enforce this in court.
 */

const haketilodb = haketilo_exports["common/indexeddb.js"];


const browser = haketilo_exports["common/browser.js"].browser;
const is_privileged_url = haketilo_exports["common/misc.js"].is_privileged_url;
const csp_header_regex = haketilo_exports["common/misc.js"].csp_header_regex;
const sha256 = haketilo_exports["common/misc.js"].sha256_async;
const decide_policy = haketilo_exports["common/policy.js"].decide_policy;

const tree = haketilo_exports["background/patterns_query_manager.js"].tree;
const default_allow = haketilo_exports["background/patterns_query_manager.js"].default_allow;

let secret;

function on_headers_received(details) {
    const url = details.url;
    if (is_privileged_url(details.url))
	return;

    let headers = details.responseHeaders;


    const policy =
	  decide_policy(tree, details.url, !!default_allow.value, secret);

    if (!policy.allow) {
	if (policy.payload)
	    headers = headers.filter(h => !csp_header_regex.test(h.name));

	headers.push({name: "Content-Security-Policy", value: policy.csp});
    }


    return {responseHeaders: headers};
}

const request_url_regex = /^[^?]*\?url=(.*)$/;
const redirect_url_template = browser.runtime.getURL("dummy") + "?settings=";

function on_before_request(details)
{
    /*
     * Content script will make a synchronous XmlHttpRequest to extension's
     * `dummy` file to query settings for given URL. We smuggle that
     * information in query parameter of the URL we redirect to.
     * A risk of fingerprinting arises if a page with script execution allowed
     * guesses the dummy file URL and makes an AJAX call to it. It is currently
     * a problem in ManifestV2 Chromium-family port of Haketilo because Chromium
     * uses predictable URLs for web-accessible resources. We plan to fix it in
     * the future ManifestV3 port.
     */
    if (details.type !== "xmlhttprequest")
	return {cancel: true};

    if (details.url.startsWith(redirect_url_template))
	return;


    /*
     * request_url should be of the following format:
     *     <url_for_extension's_dummy_file>?url=<valid_urlencoded_url>
     */
    const match = request_url_regex.exec(details.url);
    if (match) {
	const queried_url = decodeURIComponent(match[1]);

	if (details.initiator && details.initiator !== "null" &&
	    !queried_url.startsWith(details.initiator)) {
	    console.warn(`Haketilo: Blocked suspicious query of '${queried_url}' by '${details.initiator}'. This might be the result of page fingerprinting the browser.`);
	    return {cancel: true};
	}

	const policy =
	      decide_policy(tree, queried_url, !!default_allow.value, secret);
	if (!policy.error) {
	    const encoded_policy = encodeURIComponent(JSON.stringify(policy));
	    return {redirectUrl: redirect_url_template + encoded_policy};
	}
    }

    console.warn(`Haketilo: Bad request! Expected ${browser.runtime.getURL("dummy")}?url=<valid_urlencoded_url>. Got ${details.url}. This might be the result of page fingerprinting the browser.`);

    return {cancel: true};
}

const all_types = [
    "main_frame", "sub_frame", "stylesheet", "script", "image", "font",
    "object", "xmlhttprequest", "ping", "csp_report", "media", "websocket",
    "other", "main_frame", "sub_frame"
];

async function start(secret_)
{
    secret = secret_;

    const extra_opts = ["blocking", "extraHeaders"];

    browser.webRequest.onHeadersReceived.addListener(
	on_headers_received,
	{urls: ["<all_urls>"], types: ["main_frame", "sub_frame"]},
	extra_opts.concat("responseHeaders")
    );

    browser.webRequest.onBeforeRequest.addListener(
	on_before_request,
	{urls: [browser.runtime.getURL("dummy") + "*"], types: all_types},
	extra_opts
    );
}
this.haketilo_exports["background/webrequest.js"].start = (start);
}
}).call({
    haketilo_exports: this.haketilo_exports,
    haketilo_this:    this
});
